#include "main.h"

pipe_fd * pipes;
int fd_events_log;
int fd_pipes_log;
int N;

int main(int argc, char * argv[]) {
  int balances_count = get_balances_count(argc, argv);
  int *balances = get_balances(balances_count, argv);
      N = balances_count + 1;

  fd_events_log = open(evengs_log, O_WRONLY | O_APPEND | O_CREAT | O_TRUNC, 0777);
  if (fd_events_log < 0)
    log_error(OPEN_LOG_FILES_ERROR);
  
  fd_pipes_log = open(pipes_log, O_WRONLY | O_APPEND | O_CREAT | O_TRUNC, 0777);
  if (fd_pipes_log < 0)
    log_error(OPEN_LOG_FILES_ERROR);

  // open pipes
  open_pipes();
  pid_t all_pids[N];

  all_pids[0] = getpid();
  
  // fork processes
  for (local_id i = 1; i < N; i++) {
    all_pids[i] = fork();
  
    switch(all_pids[i]) {
    	case 0: // child
     	  pipe_work(i, balances[i - 1]);
   	    exit(0);
        break;
      case -1: // error
        log_error(FORK_ERROR);
        break;
      default: // parent
        break;
    }
  }

  parent_work();

  close(fd_events_log);
  close(fd_pipes_log);

  exit(0);
 }

void log_error(int err_num) {
  switch(err_num) {
    case ARGUMENTS_FORMAT_ERROR: // show ussage
      printf("%s", help_ussage_msg);
      exit(1);
      break;
    case OPEN_LOG_FILES_ERROR:
      printf("%s", open_file_log_error_msg);
      exit(1);
      break;
    case FORK_ERROR:
      printf("%s", fork_error_msg);
      exit(1);
      break;
    case UNKNOWN_MESSAGE_TYPE_ERROR:
      printf("%s", message_type_error_msg);
      exit(1);
      break;
    case SEND_MULTICAST_ERROR:
      printf("%s", send_multicast_error_msg);
      exit(1);
      break;
    case RECEIVE_MULTICAST_TYPE_ERROR:
      printf("%s", receive_multicast_type_error_msg);
      exit(1);
      break;
    case UNEXPECTED_MESSAGE_TYPE_ERROR:
      printf("%s", unexpected_message_type_error_msg);
      exit(1);
    case TRANSFER_PARAMETERS_ERROR:
      printf("%s", transfer_parameters_error_msg);
      exit(1);
    case DUPLICATED_STOP_MESSAGE_ERROR:
      printf("%s", duplicated_stop_message_error_msg);
      exit(1);
    case EARLY_BALANCE_HISTORY_REQUEST_ERROR:
      printf("%s", early_balance_history_request_error_msg);
      exit(1);
    default:
      log_error(UNKNOWN_MESSAGE_TYPE_ERROR);
      break;
    }
}

void log_event(int num_event, local_id l_id, local_id to_id, balance_t s_balance) {
  char buf[MAX_MESSAGE_LEN];

  switch(num_event) {
    case LOG_STARTED:
      sprintf(buf, log_started_fmt, 
        get_physical_time(), l_id, getpid(), getppid(), s_balance);
      break;
    case LOG_STARTED_RECIEVED_ALL:
      sprintf(buf, log_received_all_started_fmt,
        get_physical_time(), l_id);
      break;
    case LOG_DONE:
      sprintf(buf, log_done_fmt,
        get_physical_time(), l_id, s_balance);
      break;
    case LOG_TRANSFER_OUT:
      sprintf(buf, log_transfer_out_fmt, 
        get_physical_time(), l_id, s_balance, to_id);
      break;
    case LOG_TRANSFER_IN:
      sprintf(buf, log_transfer_in_fmt,
        get_physical_time(), l_id, s_balance, to_id);
      break;
    case LOG_DONE_RECIEVED_ALL:
      sprintf(buf, log_received_all_done_fmt,
        get_physical_time(), l_id);
      break;
    default:
      log_error(UNKNOWN_MESSAGE_TYPE_ERROR);
      break;
    }

  printf("%s", buf);
  write(fd_events_log, buf, strlen(buf));
}

void log_pipe(int num_event, int from, int to) {
  char buf[100];

  switch (num_event) {
    case 0:
      sprintf(buf, pipe_opend_msg, from, to);
      write(fd_pipes_log, buf, strlen(buf));
      break;
    case 1:
      sprintf(buf, pipe_closed_msg, from, to);
      write(fd_pipes_log, buf, strlen(buf));
      break;
    }
}

void open_pipes() {
  pipes = (pipe_fd *)malloc(N * (N - 1) * sizeof(pipe_fd));
  pipe_fd * tmp;

  int fds[2];

  for (int i = 0; i < N; i++)
    for (int j = 0; j < N; j++)
      	if (i != j) {
          pipe2(fds, O_NONBLOCK); 
      	  tmp = get_struct(i, j);
      	  tmp->fd_write = fds[1];
      	  log_pipe(0, i, j);

      	  tmp = get_struct(j, i);
      	  tmp->fd_read = fds[0];
      	  log_pipe(0, j, i);
      	}
}

void close_redundant_pipes(local_id id) {
  pipe_fd * tmp;

  for (local_id i = 0; i < N; i++) {
    if (i == id) 
      continue;

    for (local_id j = 0; j < N; j++) {
     	if (i != j) {
        tmp = get_struct(i, j);
    	    
        close(tmp->fd_write);	 
        close(tmp->fd_read);
        log_pipe(1, i, j);
      }
    }
  }

}

void close_self_pipes(local_id id) {
  pipe_fd * tmp;

  for (local_id i = 0; i < N; i++) {
    if (i != id) {
      tmp = get_struct(id, i);
      close(tmp->fd_write);
      close(tmp->fd_read);
      log_pipe(1, id, i);
    }
  }
}

// formirate second index in memory (shift)
// by indexes in matrix 
int f_index(int i, int j) {
  if (j >= i)
    return j - 1;
  else
    return j;
}

pipe_fd* get_struct(int from, int to) {
  if (from == to)
    return NULL;

  int       index = from * (N - 1) + f_index(from, to);
  pipe_fd   *result = pipes + (size_t)index;
  
  return result;
}

void parent_work() {  
  local_id id = 0;

  close_redundant_pipes(id);
  
  // get all STARTED messages
  if (receive_multicast(id, STARTED) < 0) 
    log_error(RECEIVE_MULTICAST_TYPE_ERROR);

  // run bank_robbery (send TRANSFER messages)
  bank_robbery(NULL, N - 1);

  // send STOP messages to all
  Message m;
  m.s_header.s_type = STOP;
  m.s_header.s_magic = MESSAGE_MAGIC;
  m.s_header.s_payload_len = 0;
  m.s_header.s_local_time = get_physical_time();

  if (send_multicast(&id, &m) == -1)
    log_error(SEND_MULTICAST_ERROR);

  // get all DONE messages
  if (receive_multicast(id, DONE) < 0)
    log_error(RECEIVE_MULTICAST_TYPE_ERROR);

  AllHistory all_history;
  get_all_history_messages(&all_history);
  
  wait_children();
  close_self_pipes(id);
  free(pipes);
  print_history(&all_history);
}

void get_all_history_messages(AllHistory * all_history) {
  Message  request;
  Message  reply;
  int      max_history_len = 0;
  local_id self = 0;

  request.s_header.s_magic        = MESSAGE_MAGIC;
  request.s_header.s_type         = BALANCE_HISTORY;
  request.s_header.s_payload_len  = 0;
  all_history->s_history_len      = N - 1;

  for (local_id i = 0; i < N - 1; i++) {
    request.s_header.s_local_time = get_physical_time();

    send(&self, i + 1, &request);
    receive(&self, i + 1, &reply);

    // unpack BalanceHistory struct
    char*     data_pointer = reply.s_payload;
    local_id  local_proc_id;
    uint8_t   current_history_len;

    memcpy(&local_proc_id, data_pointer, sizeof(local_id));
    all_history->s_history[i].s_id = local_proc_id;
    data_pointer += sizeof(local_id);
    memcpy(&current_history_len, data_pointer, sizeof(uint8_t));
    all_history->s_history[i].s_history_len = current_history_len;
    data_pointer += sizeof(uint8_t);

    if (current_history_len > max_history_len)
      max_history_len = current_history_len;

    for (int j = 0; j < all_history->s_history[i].s_history_len; j++) {
      memcpy(&all_history->s_history[i].s_history[j], data_pointer, sizeof(BalanceState));
      data_pointer += sizeof(BalanceState);
    }
  }

  // equile all lengths of history
  for (local_id i = 0; i < N - 1; i++) {
    int history_len = all_history->s_history[i].s_history_len;
    if (history_len < max_history_len) {
      BalanceState bs = all_history->s_history[i].s_history[history_len - 1];

      for (int j = history_len; j < max_history_len; j++) {
        bs.s_time = j;
        all_history->s_history[i].s_history[j] = bs;
      }

      all_history->s_history[i].s_history_len = max_history_len;      
    }
  }
}

void pipe_work(local_id id, balance_t start_balance) {
  close_redundant_pipes(id);
  synchronize_other_start(id, start_balance);
  computation(id, start_balance);
  close_self_pipes(id);
}

void synchronize_other_start(local_id id, balance_t s_balance) {
  log_event(LOG_STARTED, id, 0, s_balance);
  
  Message m;

  int len = sprintf(m.s_payload, log_started_fmt, 
            get_physical_time(), id, getpid(), getppid(), s_balance);

  m.s_header.s_magic = MESSAGE_MAGIC;
  m.s_header.s_payload_len = len;
  m.s_header.s_type = STARTED;
  m.s_header.s_local_time = get_physical_time();
   
  // send message
  if (send_multicast(&id, &m) != 0)
      log_error(SEND_MULTICAST_ERROR);
 
  // wait other message
  receive_multicast(id, m.s_header.s_type);
  log_event(LOG_STARTED_RECIEVED_ALL, id, 0, s_balance);
}

void computation(local_id id, int start_balance) {
  BalanceHistory  bh;
  BalanceState    bs;
  Message         m;
  timestamp_t     last_time = 0;
  int             done_count = 0;
  int             balance = start_balance;
  bool            isInStopState = false;
  bool            isHistoryRequired = false;

  bh.s_id = id;

  bs.s_balance = balance;
  bs.s_time = 0;
  bs.s_balance_pending_in = 0;

  bh.s_history[0] = bs;

  while(1) {
    receive_any(&id, &m); 
    //printf("ID: %d\tMessage received%d\tFrom: %d\n", id, m.s_header.s_type, by_id);

    switch(m.s_header.s_type) {
    case TRANSFER: { 
      TransferOrder to;
      timestamp_t new_time = get_physical_time();
      memcpy(&to, m.s_payload, sizeof(TransferOrder));

      if (to.s_src == id) { // this process is src
        balance -= to.s_amount;

        send(&id, to.s_dst, &m);
        log_event(LOG_TRANSFER_OUT, id, to.s_dst, to.s_amount);
      } else if (to.s_dst == id) {  // this process is dst
        balance += to.s_amount;

        Message reply;
        reply.s_header.s_magic = MESSAGE_MAGIC;
        reply.s_header.s_type = ACK;
        reply.s_header.s_payload_len = 0;
        reply.s_header.s_local_time = get_physical_time();
        
        send(&id, PARENT_ID, &reply);
        log_event(LOG_TRANSFER_IN, id, to.s_src, to.s_amount);
      } else {
        log_error(TRANSFER_PARAMETERS_ERROR);
      }

      bs.s_balance = balance;
      bs.s_time = new_time;
      bh.s_history[new_time] = bs;
      
      bs = bh.s_history[last_time];
      for (last_time++; last_time < new_time; last_time++) {
        bs.s_time = last_time;
        bh.s_history[last_time] = bs;
      }

      break;
    }
    case STOP: { 
      if (isInStopState == true)
        log_error(DUPLICATED_STOP_MESSAGE_ERROR);

      Message reply;

      int len = sprintf(reply.s_payload, log_done_fmt,
               get_physical_time(), id, balance);

      reply.s_header.s_magic = MESSAGE_MAGIC;
      reply.s_header.s_type = DONE;
      reply.s_header.s_payload_len = len;
      reply.s_header.s_local_time = get_physical_time();
      
      if (send_multicast(&id, &reply) == -1)
        log_error(SEND_MULTICAST_ERROR);
      
      isInStopState = true;
      break;
    }
    case DONE: {
      done_count++;//printf("ID: %d\tdone count incremented:%d\n", id, done_count);

      if (done_count > N - 2) {
        printf("ID: %d\tToo much done messages received: %d\n", id, done_count);
        exit(1);
      }

      if (done_count == (N - 2)) {
        log_event(LOG_DONE_RECIEVED_ALL, id, 0, 0);
        
        if ((isInStopState) && (isHistoryRequired)) {
          bh.s_history_len = last_time + 1;
          send_history_message(id, &bh);
          return;          
        } else {
          //printf("is in stop:%d\tis history req:%d\n", isInStopState, isHistoryRequired);
        }
      }
     
      break;
    }
    case BALANCE_HISTORY: {//printf("ID: %d\tbalance history required\n", id);
      if (isInStopState == true) {
        if (done_count == (N - 2)) {
          bh.s_history_len = last_time + 1;
          send_history_message(id, &bh);
          return;
        } else {
          //printf("ID: %d\tdone count: %d\tbalance history was required\n", id, done_count);
          isHistoryRequired = true;
        }
      } else {
        printf("ID: %d\tdone count: %d\tis in stop state: %d\n", id, done_count, isInStopState);
        log_error(EARLY_BALANCE_HISTORY_REQUEST_ERROR);
      }
      break;
    }
    case STARTED:
    case ACK:
      log_error(UNEXPECTED_MESSAGE_TYPE_ERROR);
      break;
    default:
      log_error(UNKNOWN_MESSAGE_TYPE_ERROR);
      break;
    }
  }
}

void send_history_message(local_id id, BalanceHistory *bh) {
    /*printf("===History of: %d===\n", id);

    for (int j = 0; j <= bh->s_history_len; j++) {
      printf("\tTime: %d\tBalance: %d\n", j, bh->s_history[j].s_balance);
    }
  */
  uint16_t pl = sizeof(balance_t) + sizeof(uint8_t) + (bh->s_history_len + 1) * sizeof(BalanceState);
  Message m;
  m.s_header.s_magic = MESSAGE_MAGIC;
  m.s_header.s_type = BALANCE_HISTORY;
  m.s_header.s_payload_len = pl;
  m.s_header.s_local_time = get_physical_time();

  memcpy(&m.s_payload, bh, pl);

  send(&id, PARENT_ID, &m);
}

int send_multicast(void * self, const Message * msg) {
  local_id l_id = *(local_id *)self;

  for (local_id i = 0; i < N; i++) {
    if (i != l_id) 
      if (send(self, i, msg) != 0) {
        return -1; 
      }
  }

  return 0;
}

int send(void * self, local_id dst, const Message * msg) {
  local_id  l_id = *(local_id *)self;
  pipe_fd   *tmp = get_struct(l_id, dst);

  int result = write(tmp->fd_write, msg, sizeof(MessageHeader) + msg->s_header.s_payload_len);
  
  if (result == -1)
    return -1;

  return 0;
}

int receive_multicast(local_id self, int16_t s_type) {
  for (local_id i = 1; i < N; i++)
      if (i != self) {
        Message   m;
        if (receive(&self, i, &m) ==  -1)
          return -2; 
        
        if (m.s_header.s_type != s_type)
          return -1;
      }

  return 0;
}

int receive (void * self, local_id from, Message * msg) {
  local_id  id = *(local_id *)self;
  pipe_fd   *tmp = get_struct(id, from);

  while(1) {
    int n = read(tmp->fd_read, &(msg->s_header), sizeof(MessageHeader));

    if (n == -1) 
      continue;

    n = read(tmp->fd_read, msg->s_payload, msg->s_header.s_payload_len);
    return 0;
  }

}

int receive_any(void * self, Message * msg) {
  local_id  self_id = *(local_id *)self;

  while (1) {
    for (local_id i = 0; i < N; i++) {
      if (i == self_id) 
        continue;

      pipe_fd*  tmp = get_struct(self_id, i);
      int       bytes_count = read(tmp->fd_read, &(msg->s_header), sizeof(MessageHeader));

      if (bytes_count <= 0) 
        continue;

      bytes_count = read(tmp->fd_read, msg->s_payload, msg->s_header.s_payload_len);

      return (int)i;
    }
  }
}

void wait_children() {
  pid_t pid;
  int   status;

  while((pid = wait(&status)) > 0) { 
    // may be logged
  }
}

void transfer(void * parent_data, local_id src, local_id dst, balance_t amount) {
  // formirate transfer order
  TransferOrder to;
  to.s_src = src;
  to.s_dst = dst;
  to.s_amount = amount;

  // formirate message
  Message m;
  Message m_received;
  m.s_header.s_magic = MESSAGE_MAGIC;
  m.s_header.s_type = TRANSFER;
  m.s_header.s_local_time = get_physical_time();
  m.s_header.s_payload_len = sizeof(TransferOrder);

  memcpy(m.s_payload, &to, sizeof(TransferOrder));

  local_id id = 0;
  send(&id, src, &m);
  receive(&id, dst, &m_received); // wait ACK message from dst
}

int get_balances_count(int argc, char * argv[]) {
  if (argc < 3 || strcmp(argv[1], p_key) != 0) 
    log_error(ARGUMENTS_FORMAT_ERROR);

  int accounts = atoi(argv[2]);

  if (accounts < 1 || argc - 3 - accounts != 0) 
    log_error(ARGUMENTS_FORMAT_ERROR);

  return accounts;
}

int* get_balances(int accounts, char * argv[]) {
  int *balances = malloc(accounts * sizeof(int));

  for (int i = 0; i < accounts; i++) {
    int val = atoi(argv[i + 3]);

    if (val < 1 || val > 99)
      log_error(ARGUMENTS_FORMAT_ERROR);

    balances[i] = val;
  }

  return balances;
}
